! WHIZARD 2.2.3 Aug 15 2014
! 
! Copyright (C) 1999-2014 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions from
!     Fabian Bach <fabian.bach@desy.de>
!     Christian Speckner <cnspeckn@googlemail.com> 
!     Christian Weiss <christian.weiss@desy.de>
!     and Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module radiation_generator

  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use os_interface
  use models
  use pdg_arrays
  use particle_specifiers
  use auto_components

  implicit none
  private

  public :: radiation_generator_init
  public :: radiation_generator_t
  
  type :: radiation_generator_t
    logical :: qcd_enabled = .false.
    logical :: qed_enabled = .false.
    type(pdg_list_t) :: pl_in, pl_out
    type(split_constraints_t) :: constraints
    integer :: n_tot
    integer :: n_out
    integer :: n_loops
    real(default) :: mass_sum
    type(model_t), pointer :: radiation_model
  contains
    procedure :: init_radiation_model => &
                      radiation_generator_init_radiation_model
    procedure :: set_n => radiation_generator_set_n
    procedure :: set_constraints => radiation_generator_set_constraints
    procedure :: generate => radiation_generator_generate
  end type radiation_generator_t


  interface radiation_generator_init
     module procedure radiation_generator_init_pdg_list
     module procedure radiation_generator_init_pdg_array
  end interface


contains

  subroutine radiation_generator_init_pdg_list &
       (generator, qcd, qed, pl_in, pl_out)
    type(radiation_generator_t), intent(inout) :: generator
    logical, intent(in), optional :: qcd, qed
    type(pdg_list_t), intent(in) :: pl_in, pl_out
    if (present (qcd))  generator%qcd_enabled = qcd
    if (present (qed))  generator%qed_enabled = qed
    generator%pl_in = pl_in
    generator%pl_out = pl_out
  end subroutine radiation_generator_init_pdg_list

  subroutine radiation_generator_init_pdg_array &
       (generator, qcd, qed, pdg_in, pdg_out)
    type(radiation_generator_t), intent(inout) :: generator
    logical, intent(in), optional :: qcd, qed
    type(pdg_array_t), intent(in), dimension(:), allocatable :: pdg_in, pdg_out
    type(pdg_list_t) :: pl_in, pl_out
    integer :: i
    call pl_in%init(size (pdg_in))    
    call pl_out%init(size (pdg_out))
    do i = 1, size (pdg_in)
       call pl_in%set (i, pdg_in(i))
    end do
    do i = 1, size (pdg_out)
       call pl_out%set (i, pdg_out(i))
    end do
    call radiation_generator_init (generator, qcd, qed, pl_in, pl_out)
  end subroutine radiation_generator_init_pdg_array

  subroutine radiation_generator_init_radiation_model (generator, os_data)
    class(radiation_generator_t), intent(inout) :: generator
    type(os_data_t), intent(in) :: os_data
    type(model_list_t) :: model_list
    type(model_t), pointer :: model
    call model_list%read_model (var_str ("SM_rad"), var_str ("SM_rad.mdl"), &
                                os_data, model)
    generator%radiation_model => model
  end subroutine radiation_generator_init_radiation_model

  subroutine radiation_generator_set_n (generator, n_in, n_out, n_loops)
    class(radiation_generator_t), intent(inout) :: generator
    integer, intent(in) :: n_in, n_out, n_loops
    generator%n_tot = n_in + n_out
    generator%n_out = n_out
    generator%n_loops = n_loops
  end subroutine radiation_generator_set_n

  subroutine radiation_generator_set_constraints &
       (generator, set_n_loop, set_mass_sum, &
        set_selected_particles, set_required_particles)
    class(radiation_generator_t), intent(inout) :: generator
    logical, intent(in) :: set_n_loop   
    logical, intent(in) :: set_mass_sum
    logical, intent(in) :: set_selected_particles
    logical, intent(in) :: set_required_particles
    integer :: i, n, n_constraints
    integer :: n_in, n_out
    type(pdg_list_t) :: pl_match
    type(pdg_array_t) :: pdg_gluon, pdg_photon

    type(pdg_array_t) :: pdg_test
    n_constraints = 1 + count([set_n_loop, set_mass_sum, &
         set_selected_particles, set_required_particles])
    associate (constraints => generator%constraints)
      n = 1
      call constraints%init (n_constraints)
      call constraints%set (n, constrain_n_tot (generator%n_tot))
      n = n+1
      if (set_n_loop) then
         call constraints%set (n, constrain_n_loop(generator%n_loops))
         n = n+1
      end if 
      if (set_mass_sum) then
        call constraints%set (n, constrain_mass_sum(generator%mass_sum))
        n = n+1
      end if
      if (set_selected_particles) then
        n_in = generator%pl_in%get_size ()
        n_out = generator%pl_out%get_size ()
        call pl_match%init (n_in+n_out+1)
        do i = 1, n_in 
           call pl_match%set (i, generator%pl_in%get (i))
        end do
        do i = 1, n_out
           call pl_match%set (i + n_in, generator%pl_out%get(i))
        end do          
        pdg_gluon = 21; pdg_photon = 22
          if (generator%qcd_enabled .and..not. generator%qed_enabled) then
             call pl_match%set (n_in + n_out, pdg_gluon)
          else if (.not. generator%qcd_enabled .and. generator%qed_enabled) then
             call pl_match%set (n_in + n_out, pdg_photon)
          else if (generator%qcd_enabled .and. generator%qed_enabled) then
             call pl_match%set (n_in + n_out, pdg_gluon // pdg_photon)
          end if
        call constraints%set (n, constrain_insert (pl_match))
        n = n+1
      end if
      if (set_required_particles) then
         !!! Dont know what to do yet
      end if
    end associate

  end subroutine radiation_generator_set_constraints

  subroutine radiation_generator_generate (generator, prt_in, prt_out)
    class(radiation_generator_t), intent(in) :: generator
    type(string_t), intent(out), dimension(:), allocatable :: prt_in, prt_out
    type(pdg_array_t) :: pdg_total
    type(pdg_array_t), dimension(:), allocatable :: pdg_tmp
    type(if_table_t) :: if_table
    type(pdg_list_t), dimension(:), allocatable :: pl_in, pl_out
    integer :: i, j, k

    allocate (pl_in (1), pl_out (1))

    pl_in(1) = generator%pl_in
    pl_out(1) = generator%pl_out

    call if_table%init &
         (generator%radiation_model, pl_in, pl_out, generator%constraints)
    call if_table%radiate (generator%constraints)

    do i = 1, if_table%get_length ()
      call if_table%get_pdg_out (i, pdg_tmp)
      if (size (pdg_tmp) == generator%n_tot) then
         call if_table%get_particle_string (i, 2, prt_in, prt_out)
      end if
    end do
  end subroutine radiation_generator_generate


end module radiation_generator

